from datetime import datetime
from deltalake import DeltaTable

import pyarrow as pa
from typing import List, Optional, Tuple, Any, Union
from bmsdna.lakeapi.core.types import FileTypes
from bmsdna.lakeapi.context.df_base import FLAVORS, ExecutionContext, ResultData, get_sql
import pyarrow.dataset
import pypika.queries
from pypika.terms import Term
import pypika.functions
import pypika.enums
import pypika
import os
from datetime import datetime, timezone
from bmsdna.lakeapi.core.config import SearchConfig
from uuid import uuid4
from adbc_driver_sqlite.dbapi import Connection, Cursor, connect


class SqliteResultData(ResultData):
    def __init__(
        self,
        original_sql: Union[pypika.queries.QueryBuilder, str],
        connection: Connection,
        chunk_size: int,
    ) -> None:
        super().__init__(chunk_size=chunk_size)
        self.original_sql = original_sql
        self.connection = connection
        self._arrow_schema = None
        self._df = None

    def columns(self):
        return self.arrow_schema().names

    def query_builder(self) -> pypika.queries.QueryBuilder:
        return pypika.Query.from_(self.original_sql)

    def arrow_schema(self) -> pa.Schema:
        if self._arrow_schema is not None:
            return self._arrow_schema
        query = get_sql(self.original_sql, limit=10)  # limit of at least 1 is needed for sqlite!
        with self.connection.cursor() as cur:
            cur.execute(query)
            self._arrow_schema = cur.fetch_arrow_table().schema
        return self._arrow_schema

    @property
    def df(self):
        if self._df is None:
            query = get_sql(self.original_sql)
            with self.connection.cursor() as cur:
                cur.execute(query)
                self._df = cur.fetch_arrow_table()
        return self._df

    def to_pandas(self):
        return self.df.to_pandas()

    def to_arrow_table(self):
        return self.df

    def to_arrow_recordbatch(self, chunk_size: int = 10000):
        query = get_sql(self.original_sql)
        return self.df.to_batches(chunk_size)


class SqliteExecutionContext(ExecutionContext):
    def __init__(self, chunk_size: int):
        super().__init__(chunk_size=chunk_size)
        self.res_con = None
        self.connections: dict[str, Connection] = dict()
        self.persistance_file_name = None
        self.len_func = "length"

    def register_arrow(self, name: str, ds: Union[pyarrow.dataset.Dataset, pyarrow.Table]):
        raise NotImplementedError("Cannot read arrow in remote sql")

    def close(self):
        pass

    def execute_sql(
        self,
        sql: Union[
            pypika.queries.QueryBuilder,
            str,
        ],
    ) -> SqliteResultData:
        # todo: get correct connection string somehow
        assert len(self.connections) == 1
        return SqliteResultData(
            sql, chunk_size=self.chunk_size, connection=self.connections[list(self.connections.keys())[0]]
        )

    def distance_m_function(self, lat1: Term, lon1: Term, lat2: Term, lon2: Term):
        raise NotImplementedError("Not implemented")

    def json_function(self, term: Term, assure_string=False):
        raise NotImplementedError(
            "Cannot convert to JSON in remote sql"
        )  # we could but sql does not support structured types anyway, so...

    def init_search(
        self,
        source_view: str,
        search_configs: list[SearchConfig],
    ):
        raise NotImplementedError("Not supported")

    def register_datasource(
        self, name: str, uri: str, file_type: FileTypes, partitions: List[Tuple[str, str, Any]] | None
    ):
        assert file_type == "sqlite"
        self.connections[name] = connect(uri)

    def list_tables(self) -> ResultData:
        return self.execute_sql("SELECT type as table_type, name from sqlite_schema where type='table'")

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        for _, v in self.connections.items():
            v.close()
        self.connections = dict()



from typing import Sequence

from cashews import cache
from fastapi import Depends, FastAPI, HTTPException, Request, Response, status
from fastapi.security import HTTPBasic, HTTPBasicCredentials

from bmsdna.lakeapi.core.config import BasicConfig, Configs, UserConfig
from bmsdna.lakeapi.core.cache import CACHE_BACKEND, CACHE_EXPIRATION_TIME_SECONDS
from datetime import timedelta

cached = cache(ttl=timedelta(hours=3))

security = HTTPBasic()

userhashmap: dict[str, str] | None = None


@cached
async def is_correct(hash: str, pwd_str: str):
    import argon2

    ph = argon2.PasswordHasher()
    return ph.verify(hash.encode("utf-8"), pwd_str.encode("utf-8"))


def add_user_middlware(app: FastAPI, basic_config: BasicConfig, users: Sequence[UserConfig]):
    @app.middleware("http")
    async def get_username(request: Request, call_next):
        import json

        credentials = await HTTPBasic(auto_error=False)(request)
        if credentials is None:
            return Response(
                status_code=401,
                headers={"WWW-Authenticate": "Basic"},
                content=json.dumps({"detail": "Not authenticated"}),
            )

        global userhashmap
        userhashmap = userhashmap or {
            ud["name"].casefold(): ud["passwordhash"] for ud in users if ud["name"]
        }  # pay attention not to include an empty user by accident

        if not credentials.username.casefold() in userhashmap.keys():
            return Response(
                status_code=401,
                headers={"WWW-Authenticate": "Basic"},
                content=json.dumps({"detail": "Incorrect email or password"}),
            )
        pwd_str = credentials.password
        hash = userhashmap[credentials.username.casefold()]

        is_correct_password = is_correct(hash, pwd_str)
        if not isinstance(is_correct_password, bool):
            is_correct_password = await is_correct_password
        if not is_correct_password:
            return Response(
                status_code=401, headers={"WWW-Authenticate": "Basic"}, content="Incorrect email or password"
            )
        request.scope["user"] = {"username": credentials.username}
        return await call_next(request)



from pydantic import BaseModel
from bmsdna.lakeapi.context.df_base import ExecutionContext
import pypika.queries
from pypika.terms import Term
from bmsdna.lakeapi.core.config import BasicConfig, Config
from bmsdna.lakeapi.core.model import GeoModel
from bmsdna.lakeapi.core.types import NearbyConfig
import pypika
from typing import cast

NearbyType = list[tuple[GeoModel, NearbyConfig]] | None  # list of config with values


def parse_lat_lon(vl: str):
    lat, lon = vl.split(",")
    return float(lat), float(lon)


def _to_geo(v: dict | GeoModel):
    return GeoModel(**v) if isinstance(v, dict) else v


def get_nearby_filter(nearby_config: list[NearbyConfig], params: BaseModel, basic_config: BasicConfig) -> NearbyType:
    nearby_dict = {c.name.lower(): c for c in nearby_config}
    v = [
        (_to_geo(v), nearby_dict[k.lower()])
        for k, v in params.model_dump(exclude_unset=True).items()
        if k.lower() in nearby_dict and v is not None
    ]
    return v if len(v) > 0 else None


def handle_nearby_request(
    context: ExecutionContext,
    config: Config,
    params: BaseModel,
    basic_config: BasicConfig,
    *,
    source_view: str,
    query: pypika.queries.QueryBuilder
):
    if config.nearby is None:
        return query
    nearbyes = get_nearby_filter(config.nearby, params, basic_config)
    if nearbyes is None:
        return query
    context.init_spatial()
    score_sum = None
    query._orderbys = []  # reset order

    for nearby_val, nearby_cfg in nearbyes:
        fn = context.distance_m_function(
            pypika.Field(nearby_cfg.lat_col),
            pypika.Field(nearby_cfg.lon_col),
            cast(Term, Term.wrap_constant(nearby_val.lat)),
            cast(Term, Term.wrap_constant(nearby_val.lon)),
        )
        query = query.select(fn.as_(nearby_cfg.name))
        query = query.where(pypika.Field(nearby_cfg.name) <= nearby_val.distance_m)
        query = query.orderby(pypika.Field(nearby_cfg.name), order=pypika.Order.asc)
    return query



import argon2
from typing import cast, Optional
import secrets
import string


def generate_strong_password():
    # define the alphabet
    letters = string.ascii_letters
    digits = string.digits
    special_chars = string.punctuation

    alphabet = letters + digits + special_chars

    # fix password length
    pwd_length = 12

    # generate a password string
    pwd = ""
    for i in range(pwd_length):
        pwd += "".join(secrets.choice(alphabet))

    # generate password meeting constraints
    while True:
        pwd = ""
        for i in range(pwd_length):
            pwd += "".join(secrets.choice(alphabet))

        if any(char in special_chars for char in pwd) and sum(char in digits for char in pwd) >= 2:
            break
    return pwd


def useradd(name: str, pwd: Optional[str], yaml_file: str):
    hasher = argon2.PasswordHasher()
    if not pwd:
        pwd = generate_strong_password()
    print(name + ": " + pwd)
    hash = hasher.hash(pwd)
    import ruamel.yaml as yml  # need to use ruaml.yaml to write yaml because it preserves comments

    yaml = yml.YAML()
    yaml.indent = 2
    with open(yaml_file, "r", encoding="utf-8") as r:
        data = yaml.load(r)
        cast(list, data["users"]).append({"name": name, "passwordhash": hash})
    with open(yaml_file, "w", encoding="utf-8") as f:
        yaml.dump(data, f)


def useradd_cli():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("name")
    parser.add_argument("--password", required=False)
    parser.add_argument("--yaml-file", default="config.yml")
    args = parser.parse_args()
    useradd(args.name, args.password, args.yaml_file)


if __name__ == "__main__":
    useradd_cli()



import yaml
from bmsdna.lakeapi.core.config import YamlData


def get_yaml(file_path) -> YamlData:
    with open(file_path, encoding="utf-8") as f:
        yaml_config = yaml.safe_load(f)
        return yaml_config



from typing import List, Literal, Optional, Type, Union

import pyarrow as pa
import pypika
import pypika.functions as fn
import pypika.queries
import pypika.terms
from deltalake import DeltaTable
from fastapi import APIRouter, BackgroundTasks, Depends, Header, HTTPException, Query, Request
from pydantic import BaseModel

from bmsdna.lakeapi.context import get_context_by_engine
from bmsdna.lakeapi.context.df_base import ExecutionContext, ResultData, get_sql
from bmsdna.lakeapi.core.config import BasicConfig, Config, Configs, CacheConfig
from bmsdna.lakeapi.core.datasource import Datasource, filter_df_based_on_params, filter_partitions_based_on_params
from bmsdna.lakeapi.core.cache import CACHE_BACKEND, CACHE_EXPIRATION_TIME_SECONDS
from bmsdna.lakeapi.core.log import get_logger
from bmsdna.lakeapi.core.model import create_parameter_model, create_response_model
from bmsdna.lakeapi.core.partition_utils import should_hide_colname
from bmsdna.lakeapi.core.response import create_response
from bmsdna.lakeapi.core.types import Engines, OutputFileType
from cashews import cache
from bmsdna.lakeapi.endpoint.endpoint_search import handle_search_request
from bmsdna.lakeapi.endpoint.endpoint_nearby import handle_nearby_request
from starlette.responses import Response

logger = get_logger(__name__)


async def get_partitions(datasource: Datasource, params: BaseModel, config: Config) -> Optional[list]:
    parts = (
        await filter_partitions_based_on_params(
            DeltaTable(datasource.uri).metadata(),
            params.model_dump(exclude_unset=True) if params else {},
            config.params or [],
        )
        if not config.datasource or config.datasource.file_type == "delta"
        else None
    )
    return parts


def remove_search_nearby(prm_dict: dict, config: Config):
    if (not config.search and not config.nearby) or len(prm_dict) == 0:
        return prm_dict
    search_nearby_cols = [s.name.lower() for s in (config.search or [])] + [
        s.name.lower() for s in (config.nearby or [])
    ]
    return {k: v for k, v in prm_dict.items() if k.lower() not in search_nearby_cols}


async def get_params_filter_expr(
    context: ExecutionContext, columns: List[str], config: Config, params: BaseModel
) -> Optional[pypika.Criterion]:
    expr = await filter_df_based_on_params(
        context,
        remove_search_nearby(params.model_dump(exclude_unset=True) if params else {}, config),
        config.params if config.params else [],
        columns,
    )
    return expr


def get_response_model(config: Config, schema: pa.Schema) -> Optional[Type[BaseModel]]:
    response_model: Optional[type[BaseModel]] = None
    try:
        response_model = create_response_model(config.tag + "_" + config.name, schema)
    except Exception as err:
        logger.warning(f"Could not get response type for f{config.route}. Error:{err}")
        response_model = None
    return response_model


def exclude_cols(columns: List[str]) -> List[str]:
    columns = [c for c in columns if not should_hide_colname(c)]
    return columns


def is_complex_type(schema: pa.Schema, col_name: str):
    f = schema.field(col_name)
    return pa.types.is_nested(f.type)


def split_csv(csv_str: str) -> list[str]:
    import csv

    reader = csv.reader([csv_str], delimiter=",", quotechar='"')
    for i in reader:
        return i
    raise ValueError("cannot happen")


_setup_caches = set()


def _setup_cache(backend: str):
    if backend not in _setup_caches:
        cache.setup(backend)
        _setup_caches.add(backend)


def create_config_endpoint(
    schema: Optional[pa.Schema],
    apimethod: Literal["get", "post"],
    config: Config,
    router: APIRouter,
    response_model: Type,
    basic_config: BasicConfig,
    configs: Configs,
):
    route = config.route

    query_model = create_parameter_model(
        schema,
        config.tag + "_" + config.name + "_" + apimethod,
        config.params,
        config.search,
        config.nearby,
        apimethod,
    )

    api_method_mapping = {
        "post": router.post(
            route,
            tags=[config.tag],
            response_model=response_model,
            operation_id=config.tag + "_" + config.name + "_" + apimethod,
            name=config.name,
        ),
        "get": router.get(
            route,
            tags=[config.tag],
            response_model=response_model,
            operation_id=config.tag + "_" + config.name + "_" + apimethod,
            name=config.name,
        ),
    }

    cache_backend = config.cache.backend or CACHE_BACKEND  # type: ignore
    _setup_cache("disk://" if cache_backend == "auto" else cache_backend)

    cache_config = config.cache

    if not cache_config:
        cache_config = CacheConfig(expiration_time_seconds=CACHE_EXPIRATION_TIME_SECONDS)

    api_method = api_method_mapping[apimethod]
    has_complex = True
    if schema is not None:
        has_complex = any((pa.types.is_struct(t) or pa.types.is_list(t) for t in schema.types))

    @api_method
    async def data(
        request: Request,
        params: query_model = (Depends() if apimethod == "get" else None),  # type: ignore
        Accept: Union[str, None] = Header(default=None),
        limit: Optional[int] = 100,
        offset: Optional[int] = 0,
        select: Union[str, None] = Query(title="$select", alias="$select", default=None, include_in_schema=False),
        distinct: bool = Query(title="$distinct", alias="$distinct", default=False, include_in_schema=False),
        engine: Engines | None = Query(title="$engine", alias="$engine", default=None, include_in_schema=False),
        format: Optional[OutputFileType] = "json",
        jsonify_complex: bool = Query(title="jsonify_complex", include_in_schema=has_complex, default=False),
        chunk_size: int | None = Query(title="$chunk_size", include_in_schema=False, default=None),
    ):  # type: ignore
        logger.debug(f"{params.model_dump(exclude_unset=True) if params else None}Union[ ,  ]{request.url.path}")

        engine = engine or config.engine or basic_config.default_engine

        logger.debug(f"Engine: {engine}")
        real_chunk_size = chunk_size or config.chunk_size or basic_config.default_chunk_size
        context = get_context_by_engine(engine, chunk_size=real_chunk_size)
        assert config.datasource is not None
        realdataframe = Datasource(
            config.version_str, config.tag, config.name, config.datasource, context, basic_config=basic_config
        )
        parts = await get_partitions(realdataframe, params, config)
        df = realdataframe.get_df(parts or None)

        expr = await get_params_filter_expr(context, df.columns(), config, params)
        base_schema = df.arrow_schema()
        new_query = df.query_builder()
        new_query = new_query.where(expr) if expr is not None else new_query
        columns = exclude_cols(df.columns())
        if select:
            columns = [
                c for c in columns if c in split_csv(select)
            ]  # split , is a bit naive, we might want to support real CSV with quotes here
        if config.datasource.exclude and len(config.datasource.exclude) > 0:
            columns = [c for c in columns if c not in config.datasource.exclude]
        if config.datasource.sortby:
            for s in config.datasource.sortby:
                new_query = new_query.orderby(
                    s.by,
                    order=pypika.Order.desc if s.direction and s.direction.lower() == "desc" else pypika.Order.asc,
                )
        if has_complex and format in ["csv", "excel", "scsv", "csv4excel"]:
            jsonify_complex = True
        if jsonify_complex:
            new_query = new_query.select(
                *[
                    pypika.Field(c)
                    if not is_complex_type(base_schema, c)
                    else context.json_function(pypika.Field(c)).as_(c)
                    for c in columns
                ]
            )
        else:
            new_query = new_query.select(*columns)

        if distinct:
            assert len(columns) <= 3  # reduce complexity here
            new_query = new_query.distinct()

        if not (limit == -1 and config.allow_get_all_pages):
            limit = 1000 if limit == -1 else limit
            new_query = new_query.offset(offset or 0).limit(limit)

        new_query = handle_search_request(
            context, config, params, basic_config, source_view=realdataframe.tablename, query=new_query
        )
        new_query = handle_nearby_request(
            context, config, params, basic_config, source_view=realdataframe.tablename, query=new_query
        )
        logger.debug(f"Query: {get_sql(new_query)}")

        df2 = context.execute_sql(new_query)

        try:
            return await create_response(
                request.url,
                format or request.headers["Accept"],
                context=context,
                sql=new_query,
                basic_config=basic_config,
                cache_config=cache_config,
                close_context=True,
            )
        except Exception as err:
            logger.error("Error in creating response", exc_info=err)
            raise HTTPException(status_code=500)



from bmsdna.lakeapi.core.env import (
    CACHE_BACKEND,
    CACHE_EXPIRATION_TIME_SECONDS,
    CACHE_JSON_RESPONSES,
    CACHE_MAX_DISK_SIZE,
    CACHE_MAX_MEMORY_SIZE,
)
from fastapi import FastAPI
import time
from cashews import cache
from cashews.commands import Command


def get_max_cache_size(disk=True):
    if disk:
        return CACHE_MAX_DISK_SIZE
    return CACHE_MAX_MEMORY_SIZE


def add_cache_middleware(app: FastAPI):
    @app.middleware("http")
    async def add_process_time_header(request, call_next):
        start_time = time.perf_counter()
        response = await call_next(request)
        process_time = time.perf_counter() - start_time
        response.headers["X-Process-Time"] = str(process_time)
        return response

    @app.middleware("http")
    async def add_from_cache_headers(request, call_next):
        with cache.detect as detector:
            response = await call_next(request)
            if request.method.lower() != "get":
                return response
            if detector.calls:
                response.headers["X-From-Cache-keys"] = ";".join(detector.calls.keys())
        return response

    @app.middleware("http")
    async def disable_middleware(request, call_next):
        if request.headers.get("X-No-Cache"):
            with cache.disabling(Command.GET):
                return await call_next(request)
        return await call_next(request)



import mimetypes
import os
import tempfile
from enum import Enum
from typing import Union
from uuid import uuid4
import polars as pl
import pyarrow as pa
from starlette.background import BackgroundTask
from starlette.datastructures import URL
from starlette.responses import FileResponse, Response, StreamingResponse
from email.utils import format_datetime, formatdate

from bmsdna.lakeapi.context.df_base import ExecutionContext, ResultData
from bmsdna.lakeapi.core.config import BasicConfig, CacheConfig
from bmsdna.lakeapi.core.log import get_logger
from bmsdna.lakeapi.core.types import OutputFileType
from cashews import cache
from bmsdna.lakeapi.core.cache import CACHE_BACKEND, CACHE_EXPIRATION_TIME_SECONDS
from datetime import timedelta
import typing
from urllib.parse import quote
from pypika.queries import QueryBuilder
from mimetypes import guess_type
from starlette.concurrency import iterate_in_threadpool

import anyio

from starlette._compat import md5_hexdigest

logger = get_logger(__name__)

cached = cache(ttl=timedelta(hours=3))


class OutputFormats(Enum):
    AVRO = 1
    CSV = 2
    CSV4EXCEL = 12
    SEMI_CSV = 11
    XLSX = 3
    HTML = 4
    ARROW_IPC = 5
    ND_JSON = 6
    PARQUET = 7
    JSON = 8
    XML = 9
    ORC = 10
    ARROW_STREAM = 14


@cached
async def parse_format(accept: Union[str, OutputFileType]) -> tuple[OutputFormats, str]:
    realaccept = accept.split(";")[0].strip().lower()
    if realaccept == "application/avro" or realaccept == "avro":
        return (OutputFormats.AVRO, ".avro")
    elif realaccept == "text/csv" or realaccept == "csv":
        return (OutputFormats.CSV, ".csv")
    elif realaccept == "text/csv+semicolon" or realaccept == "scsv":
        return (OutputFormats.SEMI_CSV, ".csv")
    elif realaccept == "x.text/csv+excel" or realaccept == "csv4excel":
        return (OutputFormats.CSV4EXCEL, ".csv")
    elif realaccept == "application/xml" or realaccept == "xml":
        return (OutputFormats.XML, ".xml")
    elif realaccept == "orc":
        return (OutputFormats.ORC, ".orc")
    elif realaccept == mimetypes.guess_type("file.xlsx") or realaccept == "xlsx":
        return (OutputFormats.XLSX, ".xlsx")
    elif realaccept == "text/html" or realaccept == "html":
        return (OutputFormats.HTML, ".html")
    elif realaccept == "application/vnd.apache.arrow.stream" or realaccept == "arrow-stream":
        return (OutputFormats.ARROW_STREAM, "")

    elif (
        realaccept == "application/x-arrow"
        or realaccept == "application/x-arrow"
        or realaccept == "application/vnd.apache.arrow.file"
        or realaccept == "arrow"
        or realaccept == "feather"
        or realaccept == "ipc"
    ):
        return (OutputFormats.ARROW_IPC, ".arrow")
    elif realaccept == "application/json+newline" or realaccept == "application/jsonl" or realaccept == "ndjson":
        return (OutputFormats.ND_JSON, ".ndjson")
    elif realaccept == "application/parquet" or realaccept == "parquet":
        return (OutputFormats.PARQUET, ".parquet")
    else:
        return (OutputFormats.JSON, ".json")


def write_frame(
    url: URL, content: ResultData, format: OutputFormats, out: str, basic_config: BasicConfig
) -> list[str]:
    if format == OutputFormats.AVRO:
        import polars as pl

        ds = pl.from_arrow(content.to_arrow_recordbatch(content.chunk_size))
        assert isinstance(ds, pl.DataFrame)
        ds.write_avro(out)
    elif format == OutputFormats.CSV:
        content.write_csv(out, separator=",")
    elif format == OutputFormats.SEMI_CSV:
        content.write_csv(out, separator=";")
    elif format == OutputFormats.CSV4EXCEL:  # need to write sep=, on first line
        content.write_csv(out + "_u8", separator=",")  # type: ignore
        with open(
            out, mode="wb"
        ) as f:  # excel wants utf-16le which polars does not support. therefore we need reencoding
            f.write(b"sep=,\n")  # add utf-8 bom at beginning
            with open(out + "_u8", mode="r", encoding="utf-8") as c8:
                line = c8.readline()
                while line != "":
                    f.write(line.encode("utf-16le"))
                    line = c8.readline()
            return [out + "_u8"]
    elif format == OutputFormats.XLSX:
        import polars as pl

        ds = pl.from_arrow(content.to_arrow_table())
        assert isinstance(ds, pl.DataFrame)
        ds.write_excel(out, autofit=True)
    elif format == OutputFormats.HTML:
        content.to_pandas().to_html(out, index=False)

    elif format == OutputFormats.XML:
        content.to_pandas().to_xml(out, index=False, parser="etree")

    elif format == OutputFormats.ARROW_IPC:
        with content.to_arrow_recordbatch(content.chunk_size) as batches:
            with pa.OSFile(out, "wb") as sink:
                with pa.ipc.new_file(sink, batches.schema) as writer:
                    for batch in batches:
                        writer.write(batch)

    elif format == OutputFormats.ARROW_STREAM:
        with content.to_arrow_recordbatch(content.chunk_size) as batches:
            with pa.OSFile(out, "wb") as sink:
                with pa.ipc.new_stream(sink, batches.schema) as writer:
                    for batch in batches:
                        writer.write_batch(batch)

    elif format == OutputFormats.ND_JSON:
        content.write_nd_json(out)
    elif format == OutputFormats.PARQUET:
        content.write_parquet(out)
    else:
        content.write_json(out)
    return []


Content = typing.Union[str, bytes]
SyncContentStream = typing.Iterator[Content]
AsyncContentStream = typing.AsyncIterable[Content]
ContentStream = typing.Union[AsyncContentStream, SyncContentStream]


class StreamingResponseWCharset(StreamingResponse):
    """StreamingResponseWCharset
    Combines file response with stream response
    """

    body_iterator: AsyncContentStream

    def __init__(
        self,
        content: ContentStream,
        status_code: int = 200,
        headers: typing.Optional[typing.Mapping[str, str]] = None,
        media_type: typing.Optional[str] = None,
        background: typing.Optional[BackgroundTask] = None,
        filename: typing.Optional[str] = None,
        stat_result: typing.Optional[os.stat_result] = None,
        method: typing.Optional[str] = None,
        content_disposition_type: str = "attachment",
        *args,
        **kwargs,
    ):
        if isinstance(content, typing.AsyncIterable):
            self.body_iterator = content
        else:
            self.body_iterator = iterate_in_threadpool(content)

        # taking over from FileResponse
        self.status_code = status_code
        self.filename = filename
        self.send_header_only = method is not None and method.upper() == "HEAD"
        if media_type is None:
            media_type = guess_type(filename or "text/pain")[0] or "text/plain"
        self.media_type = media_type
        self.background = background
        self.init_headers(headers)
        if self.filename is not None:
            content_disposition_filename = quote(self.filename)
            if content_disposition_filename != self.filename:
                content_disposition = "{}; filename*=utf-8''{}".format(
                    content_disposition_type, content_disposition_filename
                )
            else:
                content_disposition = '{}; filename="{}"'.format(content_disposition_type, self.filename)
            self.headers.setdefault("content-disposition", content_disposition)
        self.stat_result = stat_result
        if stat_result is not None:
            self.set_stat_headers(stat_result)

        if "charset" in kwargs:
            self.charset = kwargs.pop("charset")

    def set_stat_headers(self, stat_result: os.stat_result) -> None:
        content_length = str(stat_result.st_size)
        last_modified = formatdate(stat_result.st_mtime, usegmt=True)
        etag_base = str(stat_result.st_mtime) + "-" + str(stat_result.st_size)
        etag = md5_hexdigest(etag_base.encode(), usedforsecurity=False)

        self.headers.setdefault("content-length", content_length)
        self.headers.setdefault("last-modified", last_modified)
        self.headers.setdefault("etag", etag)


class TempFileWrapper:  # does not open the file which is important on windows
    def __init__(self, path: str) -> None:
        self.path = path

    @property
    def name(self):
        return self.path

    def close(self):
        os.unlink(self.path)


def get_temp_file(extension: str):
    if os.name == "nt":
        temp_file = TempFileWrapper(os.environ["TEMP"] + "/" + str(uuid4()))
    else:
        temp_file = tempfile.NamedTemporaryFile(delete=True, suffix=extension)
    return temp_file


async def create_response(
    url: URL,
    accept: str,
    context: ExecutionContext,
    sql: QueryBuilder | str,
    basic_config: BasicConfig,
    cache_config: CacheConfig,
    close_context=False,
):
    headers = {}

    format = await parse_format(accept)

    format, extension = await parse_format(accept)
    content_dispositiont_type = "attachment"
    filename = "file" + extension

    if format in [
        OutputFormats.JSON,
        OutputFormats.ND_JSON,
        OutputFormats.CSV,
        OutputFormats.SEMI_CSV,
        OutputFormats.CSV4EXCEL,
    ]:
        content_dispositiont_type = "inline"
        filename = None

    media_type = "text/csv" if extension == ".csv" else mimetypes.guess_type("file" + extension)[0]
    cache_expiration_time_seconds = cache_config.expiration_time_seconds or CACHE_EXPIRATION_TIME_SECONDS

    temp_file = get_temp_file(extension)

    def do_cache(*arg, **kwargs):
        is_cache = cache_config.cache_response and cache_expiration_time_seconds > 0
        if is_cache:
            logger.info(f"Caching key: {kwargs.get('key', None)}")
            return True
        return False

    @cache.iterator(
        ttl=cache_expiration_time_seconds,
        key="sql:{sql}:url{url}:format{format}",
        condition=do_cache,
    )
    async def response_stream(context, sql, url, format):
        chunk_size = 64 * 1024
        content = await anyio.to_thread.run_sync(context.execute_sql, sql)
        additional_files = await anyio.to_thread.run_sync(
            write_frame, url, content, format, temp_file.name, basic_config
        )

        async with await anyio.open_file(temp_file.name, mode="rb") as file:
            more_body = True
            while more_body:
                chunk = await file.read(chunk_size)
                more_body = len(chunk) == chunk_size
                yield chunk

    def clean_up():
        if close_context:
            context.close()
        try:
            temp_file.close()
        except FileNotFoundError:
            pass

    return StreamingResponseWCharset(
        content=response_stream(context, sql, url, format),
        headers=headers,
        media_type=media_type,
        content_disposition_type=content_dispositiont_type,
        filename=filename,
        background=BackgroundTask(clean_up),
        charset="utf-16le" if format == OutputFormats.CSV4EXCEL else "utf-8",
    )



import os

# In this file we have parameter that can only be set using environment variables.

KB = 1024
GB = KB ^ 3

CACHE_EXPIRATION_TIME_SECONDS = int(os.getenv("CACHE_EXPIRATION_TIME_SECONDS", 1200))
CACHE_JSON_RESPONSES = os.getenv("CACHE_JSON_RESPONSES", "1") == "1"
CACHE_BACKEND = os.getenv("CACHE_BACKEND", "auto")
CACHE_MAX_DISK_SIZE = int(os.getenv("CACHE_MAX_DISK_SIZE", GB * 10))
CACHE_MAX_MEMORY_SIZE = int(os.getenv("CACHE_MAX_MEMORY_SIZE", GB * 1))

prefixes = ["men://", "redis://", "disk://"]

if not any(CACHE_BACKEND.startswith(prefix) for prefix in prefixes):
    CACHE_BACKEND = "auto"



from bmsdna.lakeapi.context.df_base import ExecutionContext
from bmsdna.lakeapi.core.types import Engines


def get_context_by_engine(engine: Engines, chunk_size: int) -> ExecutionContext:
    match engine.lower():
        case "duckdb":
            from bmsdna.lakeapi.context.df_duckdb import DuckDbExecutionContext

            return DuckDbExecutionContext(chunk_size=chunk_size)
        case "polars":
            from bmsdna.lakeapi.context.df_polars import PolarsExecutionContext

            return PolarsExecutionContext(chunk_size=chunk_size)
        case "odbc":
            from bmsdna.lakeapi.context.df_odbc import ODBCExecutionContext

            return ODBCExecutionContext(chunk_size=chunk_size)

        case "sqlite":
            from bmsdna.lakeapi.context.df_sqlite import SqliteExecutionContext

            return SqliteExecutionContext(chunk_size=chunk_size)

        case _:
            raise Exception(f"Unknown engine {engine}")


class ExecutionContextManager:
    default_engine: Engines

    def __init__(self, default_engine: Engines, default_chunk_size: int):
        self.default_engine = default_engine
        self.contexts: dict[str, ExecutionContext] = dict()
        self.default_chunk_size = default_chunk_size

    def get_context(self, engine: Engines | None, chunk_size: int | None = None):
        real_engine: Engines = engine or self.default_engine
        if real_engine not in self.contexts:
            self.contexts[real_engine] = get_context_by_engine(real_engine, chunk_size or self.default_chunk_size)
        return self.contexts[real_engine]

    def __enter__(self, *args, **kwargs):
        return self

    def __exit__(self, *args, **kwargs):
        for _, ctx in self.contexts.items():
            ctx.__exit__(*args, **kwargs)
        self.contexts = dict()



import os
import re
from dataclasses import dataclass
from datetime import datetime
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    TypedDict,
    Union,
    cast,
    TYPE_CHECKING,
)
from typing_extensions import TypedDict, NotRequired, Required
import copy
import yaml
from bmsdna.lakeapi.core.cache import (
    CACHE_EXPIRATION_TIME_SECONDS,
    CACHE_JSON_RESPONSES,
    CACHE_BACKEND,
    CACHE_MAX_DISK_SIZE,
    CACHE_MAX_MEMORY_SIZE,
)

from bmsdna.lakeapi.core.log import get_logger
from bmsdna.lakeapi.core.partition_utils import _with_implicit_parameters
from bmsdna.lakeapi.core.types import (
    FileTypes,
    OperatorType,
    Param,
    PolaryTypeFunction,
    Engines,
    SearchConfig,
    NearbyConfig,
)
import expandvars

if TYPE_CHECKING:
    from bmsdna.lakeapi.context.df_base import ExecutionContext
logger = get_logger(__name__)


@dataclass(frozen=True)
class BasicConfig:
    enable_sql_endpoint: bool
    temp_folder_path: str
    data_path: str
    min_search_length: int
    default_engine: Engines
    default_chunk_size: int
    prepare_sql_db_hook: "Callable[[ExecutionContext], Any] | None"


def get_default_config():
    return BasicConfig(
        enable_sql_endpoint=os.getenv("ENABLE_SQL_ENDPOINT", "0") == "1",
        temp_folder_path=os.getenv("TEMP", "/tmp"),
        data_path=os.environ.get("DATA_PATH", "data"),
        min_search_length=3,
        default_engine="duckdb",
        default_chunk_size=10000,
        prepare_sql_db_hook=None,
    )


@dataclass
class Filter:
    key: str
    operator: Literal[
        "=",
        ">",
        "<",
        ">=",
        "<=",
        "<>",
        "!=",
        "==",
        "between",
        "not between",
        "in",
        "not in",
    ]
    value: Union[Any, Tuple[Any, Any]]


@dataclass
class SortBy:
    by: str
    direction: Optional[Literal["asc", "desc"]] = "asc"


@dataclass
class Column:
    name: str
    alias: Optional[str] = None

    def __post_init__(self):
        self.alias = self.alias if self.alias else self.name


def _expand_env_vars(uri: str):
    return expandvars.expandvars(uri)


@dataclass
class DatasourceConfig:
    uri: str
    file_type: FileTypes = "delta"
    select: Optional[List[Column]] = None
    exclude: Optional[List[str]] = None
    sortby: Optional[List[SortBy]] = None
    filters: Optional[List[Filter]] = None
    table_name: Optional[str] = None
    in_memory: bool = False
    cache_expiration_time_seconds: Optional[int] = CACHE_EXPIRATION_TIME_SECONDS


@dataclass
class CacheConfig:
    cache_response: bool = CACHE_JSON_RESPONSES or True
    expiration_time_seconds: Optional[int] = CACHE_EXPIRATION_TIME_SECONDS or None
    backend: Optional[str] = CACHE_BACKEND or None
    max_size: Optional[int] = CACHE_MAX_MEMORY_SIZE or None

    def __post_init__(self):
        valid_prefix = ["mem://", "disk://", "memory", "disk", "auto", "redis://"]
        if self.backend is not None:
            assert any(
                self.backend.startswith(prefix) for prefix in valid_prefix
            ), f"Backend is not valid {self.backend}. See cashews config for more info"
        if not self.backend:
            self.backend = "auto"
        if self.backend == "memory":
            self.backend = "mem://"
        if self.backend == "disk":
            self.backend = "disk://"


@dataclass
class Config:
    name: str
    tag: str
    datasource: Optional[DatasourceConfig] = None
    config_from_delta: Optional[bool] = False
    version: Optional[int] = 1
    api_method: Union[Literal["get", "post"], List[Literal["get", "post"]]] = "get"
    params: Optional[List[Union[Param, str]]] = None
    timestamp: Optional[datetime] = None
    cache_expiration_time_seconds: Optional[int] = CACHE_EXPIRATION_TIME_SECONDS
    allow_get_all_pages: Optional[bool] = False
    search: Optional[List[SearchConfig]] = None
    nearby: Optional[List[NearbyConfig]] = None
    engine: Optional[Engines] = None
    chunk_size: Optional[int] = None
    cache: Optional[CacheConfig] = None

    def __post_init__(self):
        self.version_str = (
            str(self.version) if str(self.version or 1).startswith("v") else "v" + str(self.version or 1)
        )
        self.route = "/api/" + self.version_str + "/" + self.tag + "/" + self.name

    def __repr__(self) -> str:
        kws = [f"{key}={value!r}" for key, value in self.__dict__.items()]
        return "{}({})".format(type(self).__name__, ", ".join(kws))

    def __str__(self) -> str:
        kws = [f"{key}={value!r}" for key, value in self.__dict__.items()]
        return "{}({})".format(type(self).__name__, ", ".join(kws))

    @classmethod
    def _from_dict(cls, config: Dict, basic_config: BasicConfig):
        name = config["name"]
        tag = config["tag"]
        datasource: dict[str, Any] = config.get("datasource", {})
        file_type: FileTypes = datasource.get(
            "file_type",
            "delta"
            if config.get("engine", None) not in ["odbc", "sqlite"]
            else config.get("engine", None),  # for odbc and sqlite, the only meaningful file_type is odbc/sqlite
        )
        uri = _expand_env_vars(datasource.get("uri", tag + "/" + name))
        if config.get("config_from_delta"):
            assert file_type == "delta"
            real_path = os.path.join(basic_config.data_path, uri)
            if not os.path.exists(os.path.join(real_path, "_delta_log")):
                logger.warning(f"Not a real delta path: {real_path}")
            else:
                import deltalake
                import json

                try:
                    dt = deltalake.DeltaTable(real_path)
                    cfg = json.loads(dt.metadata().configuration.get("lakeapi.config", "{}"))
                    config = config | cfg  # simple merge. in that case we expect config to be in delta mainly
                    datasource = config.get(
                        "datasource", {"uri": uri, "file_type": file_type}
                    )  # get data source again, could have select, columns etc
                except json.JSONDecodeError as err:
                    logger.warning(f"Not correct json: {real_path}\n{err}")

        version = config.get("version", 1)
        api_method = cast(Literal["post", "get"], config.get("api_method", "get"))
        params = config.get("params")
        assert isinstance(uri, str)
        columns = datasource.get("columns") if datasource else None
        select = datasource.get("select") if datasource else None
        exclude = datasource.get("exclude") if datasource else None
        sortby = datasource.get("sortby") if datasource else None
        cache_expiration_time_seconds = (
            datasource.get("cache_expiration_time_seconds", CACHE_EXPIRATION_TIME_SECONDS) if datasource else None
        )
        orderby = datasource.get("orderby") if datasource else None
        search_config = [SearchConfig(**c) for c in config["search"]] if "search" in config else None
        nearby_config = [NearbyConfig(**c) for c in config["nearby"]] if "nearby" in config else None
        logger.debug(config)

        _params: list[Param] = []
        if params:
            logger.debug(f"Parameter: {name} -> {params}")

            _params += [Param(name=param) if isinstance(param, str) else Param(**param) for param in params]

        sortby = orderby if orderby else sortby
        if sortby:
            logger.debug(f"SortBy: {name} -> {sortby}")
            sortby = [SortBy(by=s.get("by"), direction=s.get("direction")) for s in sortby]

        select = columns if columns else select

        if select:
            select = [Column(name=c.get("name"), alias=c.get("alias")) for c in select]

        _cache = config.get("cache", {})
        cache = CacheConfig(
            cache_response=_cache.get("cache_response", CACHE_JSON_RESPONSES),
            expiration_time_seconds=_cache.get("expiration_time_seconds", CACHE_EXPIRATION_TIME_SECONDS),
            backend=_cache.get("backend", CACHE_BACKEND),
            max_size=_cache.get("max_size", CACHE_MAX_DISK_SIZE),
        )

        datasource_obj = DatasourceConfig(
            uri=uri,
            file_type=file_type,
            select=select,
            exclude=exclude,
            in_memory=datasource.get("in_memory", False),
            sortby=sortby,
            table_name=datasource.get("table_name", None),
            filters=None,
            cache_expiration_time_seconds=cache_expiration_time_seconds,
        )
        new_params = _with_implicit_parameters(_params, file_type, basic_config, datasource_obj.uri)

        return cls(
            name=name,
            tag=tag,
            version=version,
            search=search_config,
            nearby=nearby_config,
            api_method=api_method,
            engine=config.get("engine", None),
            chunk_size=config.get("chunk_size", None),
            params=new_params,  # type: ignore
            allow_get_all_pages=config.get("allow_get_all_pages", False),
            datasource=datasource_obj,
            cache_expiration_time_seconds=cache_expiration_time_seconds,
            cache=cache,
        )

    @classmethod
    def from_dict(
        cls, config: Dict, basic_config: BasicConfig, table_names: List[tuple[int, str, str]]
    ) -> List["Config"]:
        name = config["name"]
        tag = config["tag"]

        if name == "*":
            uri = config.get("datasource", {}).get("uri", tag + "/*")
            assert uri.endswith("/*")
            folder = uri.rstrip("/*")
            root_folder = os.path.join(basic_config.data_path, folder)
            if not os.path.exists(root_folder):
                logger.warning("Path not existing: " + root_folder)
                return []
            else:
                ls = []
                for it in os.scandir(root_folder):
                    config_sub = copy.deepcopy(config)  # important! deepcopy required for subobjects like datasource
                    config_sub["name"] = it.name
                    config_sub["datasource"]["uri"] = config["datasource"]["uri"].replace("/*", "/" + it.name)
                    tbl_name = (config_sub.get("version", 1), config_sub["tag"], config_sub["name"])
                    file_type = config_sub["datasource"].get("file_type", "delta")
                    res_name = config_sub
                    if tbl_name not in table_names and (
                        (it.is_dir() and file_type == "delta") or (it.is_file() and file_type != "delta")
                    ):
                        ls.append(cls._from_dict(config_sub, basic_config))
            return ls
        else:
            return [cls._from_dict(config, basic_config)]

    async def to_dict(self) -> dict:
        return self.__dict__


class HttpConfig(TypedDict):
    bind: NotRequired[str]
    max_requests: NotRequired[int]
    max_requests_jitter: NotRequired[int]
    timeout: NotRequired[int]
    graceful_timeout: NotRequired[int]


class UserConfig(TypedDict):
    name: Required[str]
    passwordhash: Required[str]


class AppConfig(TypedDict):
    tile: NotRequired[str]
    description: NotRequired[str]
    version: NotRequired[str]
    static_file_path: NotRequired[str]
    logo_path: NotRequired[str]


class YamlData(TypedDict):
    tables: list[Config]
    users: list[UserConfig]
    app: AppConfig


@dataclass(frozen=True)
class Configs:
    configs: Sequence[Config]
    users: Sequence[UserConfig]

    def get_config_by_route(self, route: str):
        for config in self.configs:
            if config.route == route:
                return config

    def __iter__(self):
        for i in self.configs:
            yield i

    @classmethod
    def from_yamls(
        cls,
        basic_config: BasicConfig,
        root: str = "config",
        exclude_internal: bool = True,
        internal_pattern: str = "_",
    ):
        if re.search(r"\.ya?ml$", root):
            files = [root]
        else:
            files = [
                os.path.join(root, file)
                for file in os.listdir(root)
                if not file.startswith(internal_pattern) or not exclude_internal
            ]

        tables = []
        users = []
        for file in files:
            if file.endswith(".yml"):
                with open(file, encoding="utf-8") as f:
                    y = yaml.safe_load(f)
                    yaml_data_tables = y.get("tables")
                    if yaml_data_tables:
                        tables += [y for y in yaml_data_tables if y.get("name")]
                    yaml_data_users = y.get("users")
                    if yaml_data_users:
                        users += yaml_data_users

        flat_map = lambda f, xs: [y for ys in xs for y in f(ys)]
        table_names = [(t.get("version", 1), t["tag"], t["name"]) for t in tables if t["name"] != "*"]
        configs = flat_map(lambda x: Config.from_dict(x, basic_config=basic_config, table_names=table_names), tables)
        return cls(configs, users)



# pyright: reportUndefinedVariable=false, reportGeneralTypeIssues=false

import datetime
from typing import Any, Dict, List, Literal, Union, cast, Optional, Iterable
from bmsdna.lakeapi.context.df_base import ResultData
from cashews import cache
from fastapi import Query
from pydantic import ConfigDict, BaseModel, create_model
from pydantic.fields import FieldInfo
from bmsdna.lakeapi.context.df_base import ResultData
from bmsdna.lakeapi.core.config import Param, SearchConfig, NearbyConfig
from bmsdna.lakeapi.core.partition_utils import should_hide_colname
from bmsdna.lakeapi.core.types import OperatorType
from bmsdna.lakeapi.core.env import CACHE_EXPIRATION_TIME_SECONDS
import pyarrow as pa
from datetime import timedelta
import logging

logger = logging.getLogger(__name__)

cached = cache(ttl=timedelta(hours=3))


def _make_model(v, name):
    if type(v) is dict:
        return create_model(name, **{k: _make_model(v, k) for k, v in v.items()}), ...  # type: ignore
    return type(v), None


def make_model(v: Dict, name: str):
    return _make_model(v, name)[0]


empty_model = create_model("NoQuery")


class GeoModel(BaseModel):
    lat: float
    lon: float
    distance_m: float


_operator_postfix_map: dict[OperatorType, str] = {
    "<": "_lt",
    "<=": "_lte",
    "<>": "_ne",
    "=": "",
    ">": "_gt",
    ">=": "_gte",
    "contains": "_contains",
    "not contains": "_not_contains",
    "in": "_in",
    "not in": "_not_in",
    "not null": "_not_null",
    "null": "_null",
    "between": "_between",
    "not between": "_not_between",
    "has": "_has",
}


@cached
async def get_param_def(queryname: str, paramdef: list[Union[Param, str]]) -> Optional[tuple[Param, OperatorType]]:
    casefoldqueryname = queryname.casefold().replace(" ", "_")
    for param in paramdef:
        param = Param(name=param) if isinstance(param, str) else param
        operators = param.operators or ["="]
        for operator in operators:
            postfix = _operator_postfix_map[operator]
            if (param.name + postfix).casefold().replace(" ", "_") == casefoldqueryname:
                return (param, operator)
    return None


def get_schema_for(model_ns: str, field_name: Optional[str], field_type: pa.DataType) -> tuple[Union[type, Any], Any]:
    if pa.types.is_timestamp(field_type):
        return (Union[datetime.datetime, None], None)
    if pa.types.is_duration(field_type):
        return (Union[datetime.time, None], None)
    if pa.types.is_date(field_type):
        return (Union[datetime.date, None], None)
    if pa.types.is_time(field_type):
        return (Union[datetime.time, None], None)
    if pa.types.is_map(field_type):
        return (Union[dict, None], None)
    if pa.types.is_struct(field_type):
        st = field_type
        assert isinstance(st, pa.StructType)
        res = {
            st.field(ind).name: get_schema_for(model_ns, st.field(ind).name, st.field(ind).type)
            for ind in range(0, st.num_fields)
        }
        return (
            Union[
                create_model(model_ns + ("_" + field_name if field_name else ""), **res, __base__=TypeBaseModel),  # type: ignore
                None,
            ],
            None,
        )
    if pa.types.is_list(field_type) or pa.types.is_large_list(field_type):
        if field_type.value_type is None:
            return (Union[List[Any], None], [])

        itemtype = cast(Any, get_schema_for(model_ns, field_name, field_type.value_type)[0])
        return (Union[List[itemtype], None], [])
    if pa.types.is_integer(field_type):
        return (Union[int, None], None)
    if pa.types.is_boolean(field_type):
        return (Union[bool, None], None)
    if pa.types.is_decimal(field_type) or pa.types.is_floating(field_type):
        return (Union[float, None], None)
    if pa.types.is_string(field_type) or pa.types.is_large_string(field_type):
        return (Union[str, None], None)
    if pa.types.is_union(field_type):
        types = [
            get_schema_for(model_ns, field_type.field(ind).name, field_type.field(ind).type)
            for ind in range(0, field_type.num_fields)
        ]
        types.append(None)  # type: ignore
        return Union.__call__(*types)
    raise ValueError("Not supported")


def _get_datatype(schema: Optional[pa.Schema], name: str, *, inner: bool = False):
    if schema is None:
        return str | None
    try:
        field = schema.field(name)
        if inner and pa.types.is_list(field.type):
            return get_schema_for("prm_" + name, field.name, field.type.value_type)[0]
        return get_schema_for("prm_" + name, field.name, field.type)[0]
    except KeyError as err:
        return str | None


def _fix_space_names(query_params: dict[str, tuple[str, Any]]):
    res_dict = {}
    for key, value in query_params.items():
        if " " in key or key.startswith("_"):
            new_key = key.replace(" ", "_")
            if new_key.startswith("_"):
                new_key = "p" + new_key
            res_dict[new_key] = (value[0], FieldInfo(default=value[1], title=key))
        else:
            res_dict[key] = (value[0], FieldInfo(default=value[1], title=key))
    return res_dict


def create_parameter_model(
    schema: Optional[pa.Schema],
    name: str,
    params: Optional[Iterable[Union[Param, str]]],
    search: Optional[Iterable[SearchConfig]],
    nearby: Optional[Iterable[NearbyConfig]],
    apimethod: Literal["get", "post"],
):
    query_params: dict[str, tuple[str, Any]] = {}
    if not params and not search and not nearby:
        return empty_model
    for param in params or []:
        param = Param(name=param) if isinstance(param, str) else param
        operators = param.operators or ["="]
        if param.combi:
            if apimethod == "post":  # Only supported in POST Requets for now
                query_params[param.name] = (
                    Union[List[dict[str, Any]], None],  # type: ignore
                    Query(default=param.real_default if not param.required else ...),
                )
        else:
            realtype = (
                _get_datatype(schema, param.name) if schema is not None else str | None
            )  # well. no model, no real thing. just string
            for operator in operators:
                postfix = _operator_postfix_map[operator]

                if operator in ["in", "not in", "between", "not between"]:
                    if apimethod == "post":  # Only supported in POST Requets for now
                        query_params[param.name + postfix] = (
                            Union[List[realtype], List[dict[str, realtype]], None],  # type: ignore
                            Query(default=param.real_default if not param.required else ...),
                        )
                elif operator in ["has"] and schema is not None:
                    query_params[param.name + postfix] = (
                        _get_datatype(schema, param.name, inner=True),
                        param.real_default if not param.required else ...,
                    )
                else:
                    query_params[param.name + postfix] = (
                        realtype,
                        param.real_default if not param.required else ...,
                    )
    for sc in search or []:
        query_params[sc.name] = (
            Optional[str],
            None,
        )
    if apimethod == "post":
        for nb in nearby or []:
            query_params[nb.name] = (
                Optional[GeoModel],
                None,
            )
    query_params = _fix_space_names(query_params)
    try:
        query_model = create_model(name + "Parameter", **query_params)  # type: ignore
    except Exception as err:
        # we do not want to throw here as this fails startup which is bad
        logger.error(f"Could not create parameter model for {name}. Use empty instead")
        query_model = create_model(name + "Parameter")
    return query_model


class TypeBaseModel(BaseModel):
    model_config = ConfigDict(from_attributes=True)


def create_response_model(name: str, schema: pa.Schema) -> type[BaseModel]:
    props = {
        k: get_schema_for(name, schema.field(k).name, schema.field(k).type)
        for k in schema.names
        if not should_hide_colname(k)
    }

    return create_model(name, **props, __base__=TypeBaseModel)  # type: ignore



from pydantic import BaseModel
from bmsdna.lakeapi.context.df_base import ExecutionContext
import pypika.queries
import pypika.terms
from bmsdna.lakeapi.core.config import BasicConfig, Config
from bmsdna.lakeapi.core.types import SearchConfig

SearchesType = list[tuple[str, SearchConfig]] | None  # list of config with values


def get_searches(search_config: list[SearchConfig], params: BaseModel, basic_config: BasicConfig) -> SearchesType:
    search_dict = {c.name.lower(): c for c in search_config}
    v = [
        (v, search_dict[k.lower()])
        for k, v in params.model_dump(exclude_unset=True).items()
        if k.lower() in search_dict and v is not None and len(v) >= basic_config.min_search_length
    ]
    return v if len(v) > 0 else None


def handle_search_request(
    context: ExecutionContext,
    config: Config,
    params: BaseModel,
    basic_config: BasicConfig,
    *,
    source_view: str,
    query: pypika.queries.QueryBuilder
):
    if config.search is None:
        return query
    searches = get_searches(config.search, params, basic_config)
    if searches is None:
        return query
    context.init_search(source_view, config.search)
    score_sum = None

    for search_val, search_cfg in searches:
        score_sum = (
            context.search_score_function(source_view, search_val, search_cfg, alias=None)
            if score_sum is None
            else score_sum + context.search_score_function(source_view, search_val, search_cfg, alias=None)
        )
    assert score_sum is not None
    query = query.select(score_sum.as_("search_score"))
    query = query.where(pypika.terms.NotNullCriterion(pypika.queries.Field("search_score")))
    query._orderbys = []  # reset order
    query = query.orderby(pypika.Field("search_score"), order=pypika.Order.desc)
    return query



from typing import Literal, Tuple, cast

from fastapi import APIRouter, Depends, Request
from bmsdna.lakeapi.context import get_context_by_engine, ExecutionContext, ExecutionContextManager

from bmsdna.lakeapi.core.config import BasicConfig, Configs
from bmsdna.lakeapi.core.log import get_logger
from bmsdna.lakeapi.utils.fast_api_utils import _repeat_every


logger = get_logger(__name__)

all_lake_api_routers: list[Tuple[BasicConfig, Configs]] = []


def init_routes(configs: Configs, basic_config: BasicConfig):
    from bmsdna.lakeapi.endpoint.endpoint import (
        get_response_model,
        create_config_endpoint,
    )
    from bmsdna.lakeapi.endpoint.detail_endpoint import create_detailed_meta_endpoint
    from bmsdna.lakeapi.endpoint.sql_endpoint import create_sql_endpoint

    all_lake_api_routers.append((basic_config, configs))
    router = APIRouter()
    metadata = []
    with ExecutionContextManager(basic_config.default_engine, basic_config.default_chunk_size) as mgr:
        for config in configs:
            methods = (
                cast(list[Literal["get", "post"]], [config.api_method])
                if isinstance(config.api_method, str)
                else config.api_method
            )
            try:
                from bmsdna.lakeapi.core.datasource import Datasource

                assert config.datasource is not None
                realdataframe = Datasource(
                    config.version_str,
                    config.tag,
                    config.name,
                    config.datasource,
                    sql_context=mgr.get_context(config.engine),
                    basic_config=basic_config,
                )
                if not realdataframe.file_exists():
                    logger.warning(
                        f"Could not get response type for f{config.route}. Path does not exist:{realdataframe.uri}"
                    )
                    schema = None
                else:
                    schema = realdataframe.get_schema()
                metadata.append(
                    {
                        "name": config.name,
                        "tag": config.tag,
                        "route": config.route,
                        "methods": methods,
                        "file_type": config.datasource.file_type,
                        "uri": config.datasource.uri,
                        "version": config.version,
                        "schema": {n: str(schema.field(n).type) for n in schema.names} if schema else None,
                    }
                )

            except Exception as err:
                logger.warning(f"Could not get response type for f{config.route}. Error:{err}")
                schema = None

            response_model = get_response_model(config=config, schema=schema) if schema is not None else None
            create_detailed_meta_endpoint(
                schema=schema, config=config, configs=configs, router=router, basic_config=basic_config
            )
            for am in methods:
                create_config_endpoint(
                    apimethod=am,
                    config=config,
                    router=router,
                    response_model=response_model,
                    schema=schema,
                    basic_config=basic_config,
                    configs=configs,
                )

        @router.get(
            "/metadata",
            name="metadata",
        )
        async def get_metadata():
            return metadata

        if basic_config.enable_sql_endpoint:
            create_sql_endpoint(
                router=router,
                basic_config=basic_config,
                configs=configs,
            )

        @router.on_event("startup")
        @_repeat_every(seconds=60 * 60)  # 1 hour
        def _persist_search_endpoints() -> None:
            for config in configs:
                if config.search:
                    from bmsdna.lakeapi.core.datasource import Datasource

                    assert config.datasource is not None
                    with get_context_by_engine(basic_config.default_engine, basic_config.default_chunk_size) as ctx:
                        realdataframe = Datasource(
                            config.version_str, config.tag, config.name, config.datasource, ctx, basic_config
                        )
                        if realdataframe.file_exists():
                            ctx.init_search(realdataframe.tablename, config.search)

        return router



from deltalake import DeltaTable
from deltalake.table import MAX_SUPPORTED_READER_VERSION
from deltalake.fs import DeltaStorageHandler
import pyarrow.parquet as pq
import pyarrow
import pyarrow.fs as pa_fs
import os
import logging

logger = logging.getLogger(__name__)

from pyarrow.dataset import (
    Expression,
    FileSystemDataset,
    ParquetFileFormat,
    ParquetReadOptions,
)
from typing import List, Tuple, Union, Optional, Any


def _apply_col_mapping(dt: DeltaTable, logical_schema: pyarrow.Schema):
    meta = dt.metadata()
    schema = dt.schema()
    physical_fields = []
    for field in schema.fields:
        phys_name = field.metadata.get("delta.columnMapping.physicalName", field.name)
        existing_def = logical_schema.field(field.name)
        physical_fields.append(existing_def.with_name(phys_name))
    return pyarrow.schema(physical_fields)


def only_fixed_supported(dt: DeltaTable):
    return dt.protocol().min_reader_version == 2 and MAX_SUPPORTED_READER_VERSION == 1


def _quote(s: str):
    assert '"' not in s
    return '"' + s + '"'


def _literal(vl: Any):
    if isinstance(vl, str):
        return "'" + vl.replace("'", "''") + "'"
    if isinstance(vl, int) or isinstance(vl, float):
        return str(vl)
    return _literal(str(vl))


def get_sql_for_delta(dt: DeltaTable):
    dt.update_incremental()
    sql = "WITH files as ("
    file_selects = []

    colmaps = []
    for field in dt.schema().fields:
        colmaps.append((field.name, field.metadata.get("delta.columnMapping.physicalName", field.name)))
    phys_names = [cm[1] for cm in colmaps]
    for ac in dt.get_add_actions(flatten=True).to_pylist():
        fullpath = os.path.join(dt.table_uri, ac["path"])
        sc = pq.read_schema(fullpath)

        cols = sc.names
        cols_sqls = []
        for c in phys_names:
            if "partition_values" in ac and c in ac["partition_values"]:
                cols_sqls.append(_literal(ac["partition_values"][c]) + " AS " + _quote(c))
            elif "partition." + c in ac:
                cols_sqls.append(_literal(ac["partition." + c]) + " AS " + _quote(c))
            elif c in cols:
                cols_sqls.append(_quote(c))
            else:
                cols_sqls.append("NULL AS " + _quote(c))

        select = "SELECT " + ", ".join(cols_sqls) + " FROM read_parquet('" + fullpath + "')"
        file_selects.append(select)
    sql += "\r\n UNION ALL ".join(file_selects) + "\r\n)"
    colmapssql = ", ".join(('"' + phys_name + '" AS "' + name + '"' for name, phys_name in colmaps))
    return f"""{sql} 
     SELECT {colmapssql} from files"""


def get_pyarrow_dataset(
    dt: DeltaTable,
    partitions: Optional[List[Tuple[str, str, Any]]] = None,
    filesystem: Optional[Union[str, pa_fs.FileSystem]] = None,
    parquet_read_options: Optional[ParquetReadOptions] = None,
):
    dt.update_incremental()
    if not (dt.protocol().min_reader_version == 2 and MAX_SUPPORTED_READER_VERSION == 1):
        sc = dt.schema().to_pyarrow()
        return dt.to_pyarrow_dataset(), sc

    if not filesystem:
        filesystem = pa_fs.PyFileSystem(
            DeltaStorageHandler(
                dt._table.table_uri(),
                dt._storage_options,
            )
        )

    format = ParquetFileFormat(read_options=parquet_read_options)
    logical_schema = dt.schema().to_pyarrow()
    physical_schema = _apply_col_mapping(dt, logical_schema)
    fragments = [
        format.make_fragment(
            file,
            filesystem=filesystem,
            partition_expression=part_expression,
        )
        for file, part_expression in dt._table.dataset_partitions(physical_schema, partitions)
    ]

    schema = physical_schema

    dictionary_columns = format.read_options.dictionary_columns or set()
    if dictionary_columns:
        for index, field in enumerate(schema):
            if field.name in dictionary_columns:
                dict_field = field.with_type(pyarrow.dictionary(pyarrow.int32(), field.type))
                schema = schema.set(index, dict_field)

    fs = FileSystemDataset(fragments, schema, format, filesystem)
    return fs, logical_schema


def get_pyarrow_table(
    dt: DeltaTable,
    partitions: Optional[List[Tuple[str, str, Any]]] = None,
    filesystem: Optional[Union[str, pa_fs.FileSystem]] = None,
    parquet_read_options: Optional[ParquetReadOptions] = None,
):
    ds, real_schema = get_pyarrow_dataset(
        dt, partitions=partitions, filesystem=filesystem, parquet_read_options=parquet_read_options
    )
    t = ds.to_table().rename_columns(real_schema.names)
    return t


if __name__ == "__main__":
    d = DeltaTable("tests/data/delta/table_w_col_map")

    ds, real_schema = get_pyarrow_dataset(d)
    t = ds.to_table().rename_columns(real_schema.names)

    print(repr(t))

    lns = get_sql_for_delta(d).split("\n")
    for ln in lns:
        print(ln)



from datetime import datetime
from deltalake import DeltaTable

import pyarrow as pa
from typing import List, Optional, Tuple, Any, Union
from bmsdna.lakeapi.core.types import FileTypes
from bmsdna.lakeapi.context.df_base import FLAVORS, ExecutionContext, ResultData, get_sql
import arrow_odbc
import pyarrow.dataset
import pypika.queries
from pypika.terms import Term
import pypika.functions
import pypika.enums
import pypika
import pypika.terms
import os
from datetime import datetime, timezone
from bmsdna.lakeapi.core.config import SearchConfig
from uuid import uuid4


ENABLE_COPY_TO = os.environ.get("ENABLE_COPY_TO", "0") == "1"


def _get_temp_table_name():
    return "temp_" + str(uuid4()).replace("-", "")


arrow_odbc.enable_odbc_connection_pooling()


class BatchReaderWrap:
    def __init__(self, rdr: arrow_odbc.BatchReader):
        self.rdr = rdr

    def __enter__(self, *args, **kwargs):
        return self

    @property
    def schema(self):
        return self.rdr.schema

    def __iter__(self):
        return self.rdr.__iter__()

    def __exit__(self, *args, **kwargs):
        pass


class ODBCResultData(ResultData):
    def __init__(
        self,
        original_sql: Union[pypika.queries.QueryBuilder, str],
        connection_string: str,
        chunk_size: int,
    ) -> None:
        super().__init__(chunk_size=chunk_size)
        self.original_sql = original_sql
        self.connection_string = connection_string
        self._arrow_schema = None
        self._df = None
        self.flavor: FLAVORS = "mssql" if " for SQL Server".lower() in connection_string.lower() else "ansi"

    def columns(self):
        return self.arrow_schema().names

    def query_builder(self) -> pypika.queries.QueryBuilder:
        return pypika.Query.from_(self.original_sql)

    def arrow_schema(self) -> pa.Schema:
        if self._arrow_schema is not None:
            return self._arrow_schema
        query = get_sql(self.original_sql, limit=0, flavor=self.flavor)
        batches = arrow_odbc.read_arrow_batches_from_odbc(
            query, connection_string=self.connection_string, batch_size=self.chunk_size
        )
        assert batches is not None
        self._arrow_schema = batches.schema
        return self._arrow_schema

    @property
    def df(self):
        if self._df is None:
            query = get_sql(self.original_sql, flavor=self.flavor)
            batch_reader = arrow_odbc.read_arrow_batches_from_odbc(
                query, connection_string=self.connection_string, batch_size=self.chunk_size
            )
            assert batch_reader is not None
            self._df = pa.Table.from_batches(batch_reader, batch_reader.schema)
        return self._df

    def to_pandas(self):
        return self.df.to_pandas()

    def to_arrow_table(self):
        return self.df

    def to_arrow_recordbatch(self, chunk_size: int = 10000):
        query = get_sql(self.original_sql, flavor=self.flavor)
        res = arrow_odbc.read_arrow_batches_from_odbc(
            query, connection_string=self.connection_string, batch_size=self.chunk_size
        )
        assert res is not None
        return BatchReaderWrap(res)


class ODBCExecutionContext(ExecutionContext):
    def __init__(self, chunk_size: int):
        super().__init__(chunk_size=chunk_size)
        self.res_con = None
        self.datasources = dict()
        self.persistance_file_name = None

    def register_arrow(self, name: str, ds: Union[pyarrow.dataset.Dataset, pyarrow.Table]):
        raise NotImplementedError("Cannot read arrow in remote sql")

    def close(self):
        pass

    def execute_sql(
        self,
        sql: Union[
            pypika.queries.QueryBuilder,
            str,
        ],
    ) -> ODBCResultData:
        # todo: get correct connection string somehow
        assert len(self.datasources) == 1
        return ODBCResultData(
            sql, chunk_size=self.chunk_size, connection_string=self.datasources[list(self.datasources.keys())[0]]
        )

    def json_function(self, term: Term, assure_string=False):
        raise NotImplementedError(
            "Cannot convert to JSON in remote sql"
        )  # we could but sql does not support structured types anyway, so...

    def distance_m_function(self, lat1: Term, lon1: Term, lat2: Term, lon2: Term):
        # you have to implement haversine function yourself
        return pypika.terms.Function("f_haversine", lat1, lon1, lat2, lon2)

    def init_search(
        self,
        source_view: str,
        search_configs: list[SearchConfig],
    ):
        raise NotImplementedError("Not supported")

    def register_datasource(
        self, name: str, uri: str, file_type: FileTypes, partitions: List[Tuple[str, str, Any]] | None
    ):
        assert file_type == "odbc"
        self.datasources[name] = uri

    def list_tables(self) -> ResultData:
        return self.execute_sql("SELECT table_schema, table_name as name, table_type from information_schema.tables")

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        pass
